(load "utils.scm")

(define (ast-type x)
  (define (builtin? x)
    (case x
      ('+ #t)
      ('- #t)
      ('* #t)
      ('! #t)
      ('= #t)
      ('bool->int #t)
      ('print #t)
      (else #f)))
  (cond
   ((list? x)
    (case (car x)
      ('if 'if)
      ('let 'let)
      ('lambda 'lambda)      
      ('closure 'closure) ; only available in codegen
      ('static-string 'static-string) ; only available in codegen
      ('stack 'stack) ; only available in codegen (tag that value is passed via stack)
      (else 'app)))
   ((builtin? x) 'builtin)
   ((symbol? x) 'var)
   ((integer? x) 'int-literal)
   ((boolean? x) 'bool-literal)
   ((string? x) 'string-literal)))

(define (ast-traverse f x)
  (case (ast-type x)
    ('let `(let ,(map (lambda (x) (list (car x) (f (cadr x))))
		      (let-bindings x))
	     ,@(map f (let-body x))))
    ('app (map f x))
    ('lambda `(lambda ,(lambda-args x) ,(f (lambda-body x))))
    ('if `(if ,@(map f (cdr x))))
    ('stack `(stack ,(cadr x) ,(f (caddr x))))
    (else x)))

(define (ast-collect f x)
  (define (inner y) (ast-collect f y))
  (case (ast-type x)
    ['let (append (f x)
		  (flat-map inner (let-bindings x))
		  (flat-map inner (let-body x)))]
    ['app (append (f x)
		  (flat-map inner x))]
    ['lambda (append (f x)
		     (inner (lambda-body x)))]
    ['if (append (f x)
		 (flat-map inner (cdr x)))]
    ['stack (append (f x)
		    (inner (caddr x)))]
    [else (f x)]))

(define (ast-find p x)
  (define (inner y) (ast-find p y))
  (define (any p x) (fold-left
		     (lambda (acc y) (if acc #t (p y)))
		     #f
		     x))
  (define (either . fs)
    (if (null? fs) #f
	(if (car fs) (car fs)
	    (apply either (cdr fs)))))
		     
  (case (ast-type x)
    ['let (either (p x)
		  (any inner (let-bindings x))
		  (any inner (let-body x)))]
    ['app (either (p x)
		  (any inner x))]
    ['lambda (either (p x)
		     (inner (lambda-body x)))]
    ['if (either (p x) (any inner (cdr x)))]
    ['stack (either (p x) (inner (caddr x)))]
    [else (p x)]))

(define let-bindings cadr)
(define let-body cddr)

					; (let ([(foo a b) (foo 123 345)]) a)
					;   |
					;   v
					; (let ([a (foo~0 (foo 123 345)]
                                        ;       [b (foo~1 (foo 123 345)]) a)
(define (expand-pattern-matches program)
  (define (go x)
    (define (pattern-match binding)
      (let ([binding-name (car binding)]
	    [body (cadr binding)])
	(if (eqv? (ast-type binding-name) 'var)
	    (list (list binding-name body))
	    
	    (let* ([sum-name (car binding-name)]
		   [destructor (lambda (i) (dtor-name sum-name i))]
		   [products (cdr binding-name)]

		   [data-layouts (program-data-layouts program)]

		   [type (data-tor-type data-layouts sum-name)]

		   [sums (cdr (assoc type data-layouts))]
		   [sum (assoc sum-name sums)]
		   
		   [expected-number (length (cdr sum))])

					; assert that we only do a let pattern match on an ADT with exactly one sum
	      (when (not (= 1 (length sums)))
		(error #f (format "Cannot pattern match a ~a in a let since it has ~a possible constructors"
				  type
				  (length sums))))

					; assert that there is the correct number of bindings
	      (when (not (= (length products)
			    expected-number))
		(error #f (format "Got ~a bindings: expected ~a for ~a"
			       (length products)
			       expected-number
			       binding)))
	      
	      (flat-map (lambda (y i)
			  (pattern-match (list y `(,(destructor i) ,body))))
			products
			(range 0 (length products)))))))
    (case (ast-type x)
      ['let `(let ,(flat-map pattern-match (let-bindings x))
	       ,@(map go (let-body x)))]
      [else (ast-traverse go x)]))
  (program-map-exprs go program))

(define (lambda? x)
  (and (list? x) (eq? (car x) 'lambda)))

(define (statement-type x)
  (cond
   [(and (list? x)
	 (eqv? (car x) 'data)) 'data]
   [(and (list? x)
	 (eqv? (car x) 'define)) 'define]
   [else 'expr]))


					; (A ((foo (Int Bool))
					;     (bar (Bool)))
(define (program-data-layouts program)
  (map (lambda (x) (cons (car x) (cdr x))) ; convert to assoc list
       (map cdr (filter (lambda (x) (eqv? (statement-type x) 'data))
			program))))

(define (program-defines program)
  (filter (lambda (x) (eqv? (statement-type x) 'defines))
	  program))

(define (program-map-exprs f program)
  (map (lambda (x)
	 (case (statement-type x)
	   ['expr (f x)]
	   [else x]))
       program))

(define (program-body program)
  ; hack to have multi-expression bodies
  `(let ()
     ,@(filter (lambda (x) (eqv? (statement-type x) 'expr))
	       program)))

(define (data-tor-type data-layouts tor)
  (let* ([tors (flat-map data-tors data-layouts)]
	 [info (cadr (assoc tor tors))])
    (car info)))

					; a data tor is either a constructor or destructor for an ADT
					; data-tors returns constructors and destructors for a data-layout
					; (data A (foo Int Bool)
					;         (bar Bool))
					;        |
					;        v
					; (foo   . ((A foo constructor) . (abs Int (abs Bool A))))
					; (foo~0 . ((A foo 0)           . (abs A Int)))
					; (foo~1 . ((A foo 1)           . (abs A Bool)))
					; (bar   . ((A bar constructor) . (abs Bool A)))
					; (bar~0 . ((A bar 0)           . (abs A Bool)))
					;  ------+-------------------------------------
					;  tor   | info                 | type

(define (data-tors data-layout)
  (define (constructor-type t products)
    (fold-right (lambda (x acc) `(abs ,x ,acc)) t products))

  (define (destructor ctor-name prod-type part-type index)
    (let ([name (dtor-name ctor-name index)])
      (cons name (cons (list prod-type ctor-name index) `(abs ,prod-type ,part-type)))))
  
  (let ([type-name (car data-layout)]
        [ctors (cdr data-layout)])
    (fold-right
     (lambda (ctor acc)       
       (let* ([ctor-name (car ctor)]
	      [products (cdr ctor)]
	      
	      [maker (cons ctor-name (cons (list type-name ctor-name 'constructor) (constructor-type type-name products)))]
	      
	      [dtors (map (lambda (t i) (destructor ctor-name type-name t i))
			  products
			  (range 0 (length products)))])
	 (cons maker (append dtors acc))))
     '()
     ctors)))

					; creates a type environment for a given adt definition
(define (data-tors-type-env data-layout)
  (map (lambda (x) (cons (car x) (cddr x))) (data-tors data-layout)))

(define (dtor-name ctor-name index)
  (string->symbol
   (string-append (symbol->string ctor-name)
		  "~"
		  (number->string index))))

; for use in normalized form
(define lambda-arg caadr)
; for use elsewhere
(define lambda-args cadr)
(define lambda-body caddr)

(define (references prog)
  (ast-collect
   (lambda (x)
     (case (ast-type x)
       ['var (list x)]
       [else '()]))
   prog))

(define (graph bs)
  (define (go bs orig-bs)
    (if (null? bs)
	'(() . ())
	(let* [(bind (car bs))

	       (vert (car bind))
	       (refs (filter ; only count a reference if its a binding
		      (lambda (x) (assoc x orig-bs))
		      (references (cdr bind))))
	       (edges (map (lambda (x) (cons vert x))
			   refs))

	       (rest (if (null? (cdr bs))
			 (cons '() '())
			 (go (cdr bs) orig-bs)))
	       (total-verts (cons vert (car rest)))
	       (total-edges (append edges (cdr rest)))]
	  (cons total-verts total-edges))))
  (go bs bs))

(define (successors graph v)
  (define (go v E)
    (if (null? E)
	'()
	(if (eqv? v (caar E))
	    (cons (cdar E) (go v (cdr E)))
	    (go v (cdr E)))))
  (go v (cdr graph)))

					; takes in a graph (pair of vertices, edges)
					; returns a list of strongly connected components

					; ((x y w) . ((x . y) (x . w) (w . x))

					; =>
					; .->x->y
					; |  |
					; |  v
					; .--w

					; ((x w) (y))

					; this uses tarjan's algorithm, to get reverse
					; topological sorting for free
(define (sccs graph)
  
  (let* ([indices (make-hash-table)]
	 [lowlinks (make-hash-table)]
	 [on-stack (make-hash-table)]
	 [current 0]
	 [stack '()]
	 [result '()])

    (define (index v)
      (get-hash-table indices v #f))
    (define (lowlink v)
      (get-hash-table lowlinks v #f))

    (letrec
	([strong-connect
	  (lambda (v)
	    (begin
	      (put-hash-table! indices v current)
	      (put-hash-table! lowlinks v current)
	      (set! current (+ current 1))
	      (push! stack v)
	      (put-hash-table! on-stack v #t)

	      (for-each
	       (lambda (w)
		 (if (not (hashtable-contains? indices w))
					; successor w has not been visited, recurse
		     (begin
		       (strong-connect w)
		       (put-hash-table! lowlinks
					v
					(min (lowlink v) (lowlink w))))
					; successor w has been visited
		     (when (get-hash-table on-stack w #f)
		       (put-hash-table! lowlinks v (min (lowlink v) (index w))))))
	       (successors graph v))

	      (when (= (index v) (lowlink v))
		(let ([scc
		       (let new-scc ()
			 (let ([w (pop! stack)])
			   (put-hash-table! on-stack w #f)
			   (if (eqv? w v)
			       (list w)
			       (cons w (new-scc)))))])
		  (set! result (cons scc result))))))])
      (for-each
       (lambda (v)
	 (when (not (hashtable-contains? indices v)) ; v.index == -1
	   (strong-connect v)))
       (car graph)))
    result))
